/*-
 * Copyright (c) 2010 Lawrence Stewart <lstewart@freebsd.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/errno.h>
#include <sys/kernel.h>
#include <sys/kthread.h>
#include <sys/proc.h>
#include <sys/module.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/unistd.h>
#include <sys/malloc.h>
#include <sys/systm.h>
#include "/usr/include/sys/alq.h"
#include <sys/sbuf.h>
#include <machine/stdarg.h>

MALLOC_DECLARE(M_ALQTEST);
MALLOC_DEFINE(M_ALQTEST, "alqtest", "dynamic memory used by alqtest");

#define ENABLE	0x01
#define DISABLE	0x02

static volatile uint8_t run_writer_thread;

static char logfile[PATH_MAX] = "/tmp/alqtest.log\0";

static struct thread *alq_writer_thr = NULL;

static struct alq *theq = NULL;

volatile int count = 0;


static void
alqwrite_writer_thread(void *data)
{
	char buf[11];

	struct alq *q = (struct alq *)data;

	while (run_writer_thread) {
		bzero(buf, sizeof(buf));
		sprintf(buf, "%9d\n", count++);
		alq_write(q, buf, ALQ_WAITOK);
	}

	printf("Terminating thread\n");
	kthread_exit();
}

/*
static void
alqget_writer_thread(void *data)
{
	struct ale *tmp;

	struct alq *q = (struct alq *)data;

	while (run_writer_thread) {
		tmp = alq_get(q, ALQ_WAITOK);
		sprintf(tmp->ae_data, "%9d", count++);
		tmp->ae_data[9] = '\n';
		alq_post(q, tmp);
	}

	printf("Terminating thread\n");
	kthread_exit();
}
*/

static int
manage_test_ops(uint8_t action)
{
	int error = 0;
	int nthreads = 15;
	char buf[15];
	int i = 0;
	//struct sbuf *s = NULL;

	/* init an autosizing sbuf that initially holds 200 chars */
	//if ((s = sbuf_new(NULL, NULL, 200, SBUF_AUTOEXTEND)) == NULL)
	//	return -1;

	if (action == ENABLE) {
		
		alq_open(&theq, logfile, curthread->td_ucred, 0600, 10, 20);

		run_writer_thread = 1;

		for (i = 0; i < nthreads; i++) {
			sprintf(buf, "alq_writer_thr %d", i);
			kthread_add(&alqwrite_writer_thread, theq, NULL,
			    &alq_writer_thr, RFNOWAIT, 0, buf);
			printf("Started thread: %s\n", buf);
			bzero(buf, sizeof(buf));
		}
	}
	else if (action == DISABLE && alq_writer_thr != NULL) {
		/* Tell the test threads to exit now. */
		run_writer_thread = 0;
		alq_writer_thr = NULL;
		pause("writer_wait", 5000);
		alq_close(theq);
		theq = NULL;
	}

	return (error);
}

static int
deinit(void)
{
	manage_test_ops(DISABLE);
	return 0;
}

static int
init(void)
{
	manage_test_ops(ENABLE);
	return 0;
}

/*
 * This is the function that is called to load and unload the module.
 * When the module is loaded, this function is called once with
 * "what" == MOD_LOAD
 * When the module is unloaded, this function is called twice with
 * "what" = MOD_QUIESCE first, followed by "what" = MOD_UNLOAD second
 * When the system is shut down e.g. CTRL-ALT-DEL or using the shutdown command,
 * this function is called once with "what" = MOD_SHUTDOWN
 * When the system is shut down, the handler isn't called until the very end
 * of the shutdown sequence i.e. after the disks have been synced.
 */
static int alqtest_load_handler(module_t mod, int what, void *arg)
{
	switch(what) {
		case MOD_LOAD:
			return init();
			break;

		case MOD_QUIESCE:
		case MOD_SHUTDOWN:
			return deinit();
			break;

		case MOD_UNLOAD:
			return 0;
			break;

		default:
			return EINVAL;
			break;
	}
}

/* basic module data */
static moduledata_t alqtest_mod =
{
	"alqtest",
	alqtest_load_handler, /* execution entry point for the module */
	NULL
};

/*
 * Param 1: name of the kernel module
 * Param 2: moduledata_t struct containing info about the kernel module
 *          and the execution entry point for the module
 * Param 3: From sysinit_sub_id enumeration in /usr/include/sys/kernel.h
 *          Defines the module initialisation order
 * Param 4: From sysinit_elem_order enumeration in /usr/include/sys/kernel.h
 *          Defines the initialisation order of this kld relative to others
 *          within the same subsystem as defined by param 3
 */
DECLARE_MODULE(alqtest, alqtest_mod, SI_SUB_SMP, SI_ORDER_ANY);
MODULE_DEPEND(alqtest, alq, 1, 1, 1);


